/*
 *  TAP-Win32 -- A kernel driver to provide virtual tap device functionality
 *               on Windows.  Originally derived from the CIPE-Win32
 *               project by Damion K. Wilson, with extensive modifications by
 *               James Yonan.
 *
 *  All source code which derives from the CIPE-Win32 project is
 *  Copyright (C) Damion K. Wilson, 2003, and is released under the
 *  GPL version 2 (see below).
 *
 *  All other source code is Copyright (C) James Yonan, 2003-2004,
 *  and is released under the GPL version 2 (see below).
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program (see the file COPYING included with this
 *  distribution); if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <windows.h>
#include <ddk/ntapi.h> 
#include <ddk/winddk.h>
#include <ddk/ntddk.h>


//TAP Input/Output Control
#define TAP_CONTROL_CODE(request,method) CTL_CODE (FILE_DEVICE_UNKNOWN, request, method, FILE_ANY_ACCESS)
#define TAP_IOCTL_GET_MAC               TAP_CONTROL_CODE (1, METHOD_BUFFERED)
#define TAP_IOCTL_GET_VERSION           TAP_CONTROL_CODE (2, METHOD_BUFFERED)
#define TAP_IOCTL_GET_MTU               TAP_CONTROL_CODE (3, METHOD_BUFFERED)
#define TAP_IOCTL_GET_INFO              TAP_CONTROL_CODE (4, METHOD_BUFFERED)
#define TAP_IOCTL_CONFIG_POINT_TO_POINT TAP_CONTROL_CODE (5, METHOD_BUFFERED)
#define TAP_IOCTL_SET_MEDIA_STATUS      TAP_CONTROL_CODE (6, METHOD_BUFFERED)
#define TAP_IOCTL_CONFIG_DHCP_MASQ      TAP_CONTROL_CODE (7, METHOD_BUFFERED)
#define TAP_IOCTL_GET_LOG_LINE          TAP_CONTROL_CODE (8, METHOD_BUFFERED)
#define TAP_IOCTL_CONFIG_DHCP_SET_OPT   TAP_CONTROL_CODE (9, METHOD_BUFFERED)


//Registry keys
#define ADAPTER_KEY "SYSTEM\\CurrentControlSet\\Control\\Class\\{4D36E972-E325-11CE-BFC1-08002BE10318}"
#define NETWORK_CONNECTIONS_KEY "SYSTEM\\CurrentControlSet\\Control\\Network\\{4D36E972-E325-11CE-BFC1-08002BE10318}"


//Filesystem prefixes
#define USERMODEDEVICEDIR "\\\\.\\Global\\"
#define SYSDEVICEDIR      "\\Device\\"
#define USERDEVICEDIR     "\\DosDevices\\Global\\"
#define TAPSUFFIX         ".tap"


//TAP_COMPONENT_ID defines the TAP driver. Different component IDs can reside in the system simultaneously.
#define TAP_COMPONENT_ID "tap0801"
#define TAP_ASYNCHRONOUS_WRITES 1
#define TAP_BUFFER_SIZE 4096
#define TAP_MAX_BUFFER_COUNT 64

// Data storage buffer
typedef struct tap_buffer_s 
{
    unsigned char buffer [TAP_BUFFER_SIZE];
    unsigned long read_size;
    struct tap_buffer_s* next;
} tap_buffer_t;

//Extended Overlapped Structure with different handles
//Conatains also the Win32 Structures: CRITICAL_SECTION and OVERLAPPED 
//used for async. reading/writing and Queue used for storage of read buffer content.
typedef struct tap_win32_overlapped 
{
    HANDLE handle;
    HANDLE read_event;
    HANDLE write_event;
    HANDLE output_queue_semaphore;
    HANDLE free_list_semaphore;
    CRITICAL_SECTION output_queue_cs;
    CRITICAL_SECTION free_list_cs;
    OVERLAPPED read_overlapped;
    OVERLAPPED write_overlapped;
    tap_buffer_t buffers[TAP_MAX_BUFFER_COUNT];
    tap_buffer_t* free_list;
	tap_buffer_t* output_queue_front;
    tap_buffer_t* output_queue_back;
} tap_win32_overlapped_t;

static tap_win32_overlapped_t tap_overlapped;

//Consuming a fresh empty buffer (queue unit) from "empty-list"
static tap_buffer_t* get_next_free_buffer(tap_win32_overlapped_t* const overlapped) 
{
    //Initalizing an empty Buffer List as return value
	tap_buffer_t* buffer = NULL;

	//Asynchr. I/O
    WaitForSingleObject(overlapped->free_list_semaphore, INFINITE);
    
	//Entering the Critical Section
	EnterCriticalSection(&overlapped->free_list_cs);
    
	//Getting the buffer structure from "free-list" into return value (new List)
	buffer = overlapped->free_list;
    
	//Temporary storing the "free-list" into next member of the list
	overlapped->free_list = buffer->next;
    
	//Leaving Critical Section
    LeaveCriticalSection(&overlapped->free_list_cs);

	//Singal to other threads, that accessing is done
	ReleaseSemaphore(overlapped->free_list_semaphore, 1, NULL);
    
	//Clearing the consumed buffer from "free-list"
	buffer->next = NULL;

    return buffer;
}

//Storing used buffer (new free buffer) into "free-list"
static void put_free_buffer(tap_win32_overlapped_t* const overlapped, tap_buffer_t* const buffer)
{
    // Entering critical section
	EnterCriticalSection(&overlapped->free_list_cs);
	//Store free list into next member of the list temporary
	buffer->next = overlapped->free_list;
   //Free list is not  
    overlapped->free_list = buffer;
	//leaving critical section
	LeaveCriticalSection(&overlapped->free_list_cs);
	//Singal to other threads, that accessing is done
	ReleaseSemaphore(overlapped->free_list_semaphore, 1, NULL);
}

static tap_buffer_t* get_buffer_from_output_queue(tap_win32_overlapped_t* const overlapped, const int block) 
{
    tap_buffer_t* buffer = NULL;
    DWORD result, timeout = block ? INFINITE : 0L;

    // Asynchr. I/O 
    result = WaitForSingleObject(overlapped->output_queue_semaphore, timeout); 

    switch (result) 
    { 
        // The semaphore object was signaled.
        case WAIT_OBJECT_0: 
            
            EnterCriticalSection(&overlapped->output_queue_cs);

            buffer = overlapped->output_queue_front;
            overlapped->output_queue_front = buffer->next;

            if(overlapped->output_queue_front == NULL) 
			{
              overlapped->output_queue_back = NULL;
            }

            LeaveCriticalSection(&overlapped->output_queue_cs);
            //Singal to other threads, that accessing is done
	        ReleaseSemaphore(overlapped->free_list_semaphore, 1, NULL);
            break; 

        //Semaphore was nonsignaled, so a time-out occurred.
        case WAIT_TIMEOUT:
          printf("timeout\n"); 
            
            break; 
       	case WAIT_ABANDONED:
            printf("abandoned\n"); 
            break;
            
        case WAIT_FAILED:
            printf("wait failed\n"); 
            break;         
    }
    return buffer;
}

static tap_buffer_t* get_buffer_from_output_queue_immediate(tap_win32_overlapped_t* const overlapped) 
{
    return get_buffer_from_output_queue(overlapped, 0);
}

static void put_buffer_on_output_queue(tap_win32_overlapped_t* const overlapped, tap_buffer_t* const buffer)
{
    EnterCriticalSection(&overlapped->output_queue_cs);

    if(overlapped->output_queue_front == NULL && overlapped->output_queue_back == NULL) 
	{
      overlapped->output_queue_front = overlapped->output_queue_back = buffer;
    }
	else 
	{
	  buffer->next = NULL;
      overlapped->output_queue_back->next = buffer;
      overlapped->output_queue_back = buffer;
    }
     LeaveCriticalSection(&overlapped->output_queue_cs);
     ReleaseSemaphore(overlapped->output_queue_semaphore, 1, NULL);
}


typedef struct eth_header 
{
	DWORD dest;
	DWORD source;
	WORD  type;
} eth_header_t;



static int is_tap_win32_dev(const char *guid)
{
    HKEY netcard_key;
    LONG status;
    DWORD len;
    int i = 0;

    status = RegOpenKeyEx(HKEY_LOCAL_MACHINE, ADAPTER_KEY,0,KEY_READ,&netcard_key);

    if (status != ERROR_SUCCESS) 
	{
        return FALSE;
    }

    for (;;) 
	{
        char enum_name[256];
        char unit_string[256];
        HKEY unit_key;
        char component_id_string[] = "ComponentId";
        char component_id[256];
        char net_cfg_instance_id_string[] = "NetCfgInstanceId";
        char net_cfg_instance_id[256];
        DWORD data_type;

        len = sizeof (enum_name);
        status = RegEnumKeyEx(netcard_key,i,enum_name,&len,NULL,NULL,NULL,NULL);

        if(status == ERROR_NO_MORE_ITEMS)
		{
          break;
		}
        else if(status != ERROR_SUCCESS) 
		{
          return FALSE;
        }

        snprintf (unit_string, sizeof(unit_string), "%s\\%s",
                  ADAPTER_KEY, enum_name);

        status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,unit_string,0,KEY_READ,&unit_key);

        if (status != ERROR_SUCCESS) 
		{
            return FALSE;
        } 
		else 
		{
            len = sizeof (component_id);
            status = RegQueryValueEx(unit_key,component_id_string,NULL,&data_type,component_id,&len);

            if (!(status != ERROR_SUCCESS || data_type != REG_SZ)) 
			{
                len = sizeof (net_cfg_instance_id);
                status = RegQueryValueEx(unit_key,net_cfg_instance_id_string,NULL,&data_type,net_cfg_instance_id,&len);

                if (status == ERROR_SUCCESS && data_type == REG_SZ) 
				{
                    if (!strcmp (net_cfg_instance_id, guid)) 
					{
                      RegCloseKey (unit_key);
                      RegCloseKey (netcard_key);
                      return TRUE;
                    }
                }
            }
            RegCloseKey (unit_key);
        }
        ++i;
    }

    RegCloseKey (netcard_key);
    return FALSE;
}



static int get_device_guid(char *name,int name_size,char *actual_name,int actual_name_size)
{
    LONG status;
    HKEY control_net_key;
    DWORD len;
    int i = 0;
    int stop = 0;

    status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,NETWORK_CONNECTIONS_KEY,0,KEY_READ,&control_net_key);
    
	if(status != ERROR_SUCCESS) 
	{
      return -1;
    }

    while (!stop)
    {
        char enum_name[256];
        char connection_string[256];
        HKEY connection_key;
        char name_data[256];
        DWORD name_type;
        const char name_string[] = "Name";

        len = sizeof (enum_name);
        status = RegEnumKeyEx(control_net_key,i,enum_name,&len,NULL,NULL,NULL,NULL);

        if (status == ERROR_NO_MORE_ITEMS)
		{ 
          break;
		}
        else if (status != ERROR_SUCCESS) 
		{
          return -1;
        }

        snprintf(connection_string, sizeof(connection_string),"%s\\%s\\Connection",NETWORK_CONNECTIONS_KEY, enum_name);
        status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,connection_string,0,KEY_READ,&connection_key);
        
        if(status == ERROR_SUCCESS) 
		{
            len = sizeof (name_data);
            status = RegQueryValueEx(connection_key,name_string,NULL,&name_type,name_data,&len); 

            if(status != ERROR_SUCCESS || name_type != REG_SZ) 
			{
              return -1;
            }
            else 
			{
                if(is_tap_win32_dev(enum_name)) 
				{
                    snprintf(name, name_size, "%s", enum_name);
                    if(actual_name) 
					{
                        if(strcmp(actual_name, "") != 0) 
						{
                            if(strcmp(name_data, actual_name) != 0) 
							{
                                RegCloseKey (connection_key);
                                ++i;
                                continue;
                            }
                        }
                        else 
						{
                            snprintf(actual_name, actual_name_size, "%s", name_data);
                        }
                    }
                    stop = 1;
                }
            }

            RegCloseKey (connection_key);
        }
        ++i;
    }

    RegCloseKey (control_net_key);

    if(stop == 0)
	{
      return -1;
	}

    return 0;
}



int tap_win32_set_status(HANDLE handle, int status)
{
    unsigned long len = 0;
	
	return DeviceIoControl(handle, TAP_IOCTL_SET_MEDIA_STATUS, &status, sizeof(status), &status, sizeof(status), &len, NULL);
}

void tap_win32_overlapped_init(tap_win32_overlapped_t* const overlapped, const HANDLE handle)
{
    overlapped->handle = handle;

    overlapped->read_event = CreateEvent(NULL, FALSE, FALSE, NULL);
    overlapped->write_event = CreateEvent(NULL, FALSE, FALSE, NULL);

    overlapped->read_overlapped.Offset = 0;
    overlapped->read_overlapped.OffsetHigh = 0;
    overlapped->read_overlapped.hEvent = overlapped->read_event;

    overlapped->write_overlapped.Offset = 0;
    overlapped->write_overlapped.OffsetHigh = 0;
    overlapped->write_overlapped.hEvent = overlapped->write_event;

    InitializeCriticalSection(&overlapped->output_queue_cs);
    InitializeCriticalSection(&overlapped->free_list_cs);

	//Creating unnamed Semaphore for accessing output queue
    overlapped->output_queue_semaphore = CreateSemaphore( NULL,0,TAP_MAX_BUFFER_COUNT,NULL);

    if(!overlapped->output_queue_semaphore)  
	{
      fprintf(stderr, "error creating output queue semaphore!\n");
    }

    //Creating unnamed Semaphore for accessing Free bufferlist
    overlapped->free_list_semaphore = CreateSemaphore( NULL,TAP_MAX_BUFFER_COUNT,TAP_MAX_BUFFER_COUNT,NULL);  

    if(!overlapped->free_list_semaphore)
	{
      fprintf(stderr, "error creating free list semaphore!\n");
    }

    overlapped->free_list = overlapped->output_queue_front = overlapped->output_queue_back = NULL;
    {
      unsigned index;
      for(index = 0; index < TAP_MAX_BUFFER_COUNT; index++) 
      {
        tap_buffer_t* element = &overlapped->buffers[index];
        element->next = overlapped->free_list;
        overlapped->free_list = element;
      }
    }
}

int tap_win32_write(int *handle, void *buffer, unsigned long size)
{
    tap_win32_overlapped_t *overlapped;
    unsigned long write_size;
    BOOL result;
    DWORD error;

    
    // Asynchr. I/O
    overlapped = (tap_win32_overlapped_t *)handle;
    // Asynchr. I/O
    result = GetOverlappedResult(overlapped->handle, &overlapped->write_overlapped,&write_size, FALSE);

    
    // Asynchr. I/O
    if(!result && GetLastError() == ERROR_IO_INCOMPLETE)
	{
	  WaitForSingleObject(overlapped->write_event, INFINITE);
	}
    
	// Wiriting to TAP device
	result = WriteFile(overlapped->handle, buffer, size,&write_size, &overlapped->write_overlapped);
    
    if(!result)
	{ 
	  // Asynchr. I/O
      if((error = GetLastError()) == ERROR_IO_PENDING)
      {  
        #ifndef TAP_ASYNCHRONOUS_WRITES
        WaitForSingleObject(overlapped->write_event, INFINITE);
        #endif
	  }
	  else
	  {
        return -1;
      } 
    }

    return 0;
}

// Contains the "reading from TAP" method
static DWORD WINAPI tap_win32_thread_entry(LPVOID param)
{
    tap_win32_overlapped_t *overlapped = (tap_win32_overlapped_t*)param;
    unsigned long read_size;
    BOOL result;
    DWORD dwError;
    tap_buffer_t* buffer = get_next_free_buffer(overlapped);


    for (;;) 
	{
        // Reading from TAP Interface
        result = ReadFile(overlapped->handle,buffer->buffer,sizeof(buffer->buffer),&read_size,&overlapped->read_overlapped);
        if (!result) 
		{
          dwError = GetLastError();
          if (dwError == ERROR_IO_PENDING) 
		  {
            //Asynchr. I/O
            WaitForSingleObject(overlapped->read_event, INFINITE);
            
			result = GetOverlappedResult(overlapped->handle, &overlapped->read_overlapped,&read_size, FALSE);
            
			if(!result) 
			{
               #if DEBUG_TAP_WIN32
                  LPVOID lpBuffer;
                  dwError = GetLastError();
                  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,NULL,dwError, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) & lpBuffer, 0, NULL );
                  fprintf(stderr, "Tap-Win32: Error GetOverlappedResult %d - %s\n", dwError, lpBuffer);
                  LocalFree(lpBuffer );
               #endif
            }
           } 
		   else 
		   {
             #if DEBUG_TAP_WIN32
                LPVOID lpBuffer;
                FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,NULL, dwError, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) & lpBuffer, 0, NULL );
                fprintf(stderr, "Tap-Win32: Error ReadFile %d - %s\n", dwError, lpBuffer);
                LocalFree( lpBuffer );
             #endif
           }
        }

        // Storing read packets on the output queue 
        // Taking next free buffer
        if(read_size > 0) 
		{
            buffer->read_size = read_size;
            put_buffer_on_output_queue(overlapped, buffer);
            buffer = get_next_free_buffer(overlapped);
        }
    }

    return 0;
}

// Gets buffer content from ReadingQueue
int tap_win32_read(int handle, char **pbuf, int max_size)
{
    tap_win32_overlapped_t *overlapped = (tap_win32_overlapped_t *)handle;
    int size = 0;

    tap_buffer_t* buffer = get_buffer_from_output_queue(overlapped,1);
    
    if(buffer != NULL) 
	{
      *pbuf = buffer->buffer;
      size = (int)buffer->read_size;
      
	  if(size > max_size) 
	  {
        size = max_size;
      }
    }

    return size;
}

// Clears the buffer content and stores it into "free-buffer" list
void tap_win32_free_buffer(int handle, char* pbuf) 
{
    tap_win32_overlapped_t *overlapped = (tap_win32_overlapped_t *)handle;
    tap_buffer_t* buffer = (tap_buffer_t*)pbuf;
    put_free_buffer(overlapped, buffer);
}


// Opens Handle for TAP interface with preffered Name
int tap_win32_open(int *phandle, char *prefered_name)
{
    char device_path[256];
    char device_guid[0x100];
    int device_id;
    HANDLE handle;
    BOOL res_dev_ctrl;
    char name_buffer[0x100] = {0, };
    
	struct 
	{
      unsigned long major;
      unsigned long minor;
      unsigned long debug;
    }version;

    LONG version_len;
    DWORD idThread;
    HANDLE hThread;

    if(prefered_name != NULL)
	{
      snprintf(name_buffer, sizeof(name_buffer), "%s", prefered_name);
	}
   
	if(device_id = get_device_guid(device_guid, sizeof(device_guid), name_buffer, sizeof(name_buffer)))
	{
      return -1;
	}

    snprintf(device_path, sizeof(device_path), "%s%s%s",USERMODEDEVICEDIR,device_guid,TAPSUFFIX);

    //Opens a TAP Device
	handle = CreateFile (device_path,GENERIC_READ | GENERIC_WRITE,0,0,OPEN_EXISTING,FILE_ATTRIBUTE_SYSTEM | FILE_FLAG_OVERLAPPED,0);

    if(handle == INVALID_HANDLE_VALUE) 
	{
      return -1;
    }

	res_dev_ctrl = DeviceIoControl(handle, TAP_IOCTL_GET_VERSION, &version, sizeof (version), &version, sizeof (version), &version_len, NULL);

    if(res_dev_ctrl == FALSE) 
	{
      printf("error \r\n");  
      CloseHandle(handle);
      return -1;
    }

    if(!tap_win32_set_status(handle, TRUE)) 
	{
      return -1;
    }
    
	//Initialisation of TAP
    tap_win32_overlapped_init(&tap_overlapped, handle);
    
	//Pointer to handle 
    *phandle = (int)&tap_overlapped;

    //printf("creating hThread \r\n");

    //Creating Thread with appropriate handles.
	hThread = CreateThread(NULL, 0, tap_win32_thread_entry,(LPVOID)&tap_overlapped, 0, &idThread);
    //Set high Thread priority (Time Critical)
	SetThreadPriority(hThread,THREAD_PRIORITY_TIME_CRITICAL);
	
    return 0;
}


//Close Handle for TAP INterface
int tap_win32_close(HANDLE handle) 
{
   return (int)CloseHandle(handle);
}



//*********** TEST METHODS **************************************

// Test Method. Not usable for TAP
void readPacket(int handle) 
{
	char buffer[TAP_BUFFER_SIZE];
	//printf("reading\r\n");
	int result = tap_win32_read(handle,buffer,TAP_BUFFER_SIZE);
	//printf("Result : %d \r\n", result);		
}

// Test Method. Not usable for TAP
void writePacket(int *handle) {
	char buffer[TAP_BUFFER_SIZE];
	//printf("writing\r\n");
	int result = tap_win32_write(handle,buffer,TAP_BUFFER_SIZE);
	//printf("Result : %d \r\n", result);

}

// Test Method. Not usable for TAP
void write_loop(int handle) 
{
   char buffer[9216];
   DWORD number_write;
   int result;
   int result1;
   eth_header_t *header=buffer;
   
   
  for(;;) 
  {
		
		result = tap_win32_write(handle, buffer, 9126);
		result1 = tap_win32_read(handle,buffer,9126);
		if (!result) 
		{
			printf("error while writing \r\n");
		} else 
		{
			printf("write packet, size=%d type: %x \r\n",result,header->type);	
        }
  }
}
//*********** TEST METHODS **************************************
