package ch.unibe.iam.mcast.core.java;

import java.nio.ByteBuffer;

public class CircularMessageBuffer {

	private final ByteBuffer buffer;

	private int readPosition;

	private int unread;

	public CircularMessageBuffer(int size) {
		this.buffer=ByteBuffer.allocate(size);
	}
	

	public CircularMessageBuffer(ByteBuffer buffer) {
		this.buffer = buffer;
		reset();
		this.readPosition = 0;
		this.unread = 0;
	}
	

	public void add(ByteBuffer source) {
		
		
		int size=getObjectSize(source);
		
//		System.out.println("adding object with size "+size+" buffer space "+source.remaining());
//		System.out.println("before: readPosition: "+readPosition+", unread: "+unread);
		
		if (size > buffer.capacity()) {
			throw new RuntimeException("trying to write object of size ("+size+") into a buffer with capacity "+buffer.capacity());			
		}
		
		if (source.remaining() < size) {
			throw new RuntimeException("source object is not complete in the source buffer. size="+size+", remaining: "+source.remaining());
		}
		reset();		
		// make room for the object
		while (remainingBytes() < size) {
			removeNextObject();
		}
		
		int oldLimit=source.limit();

		int totalLimit=source.position()+size;
		
		int writePosition=(readPosition+unread) % buffer.capacity();
		
		if (writePosition+size > buffer.capacity()) {
			int partSize=buffer.capacity()-writePosition;
			source.limit(source.position()+partSize);
			buffer.position(writePosition);
			buffer.limit(buffer.capacity());
			buffer.put(source);
			unread+=partSize;
			writePosition=0;
		}
		
		source.limit(totalLimit);
		buffer.position(writePosition);
		int remaining = source.remaining();
		buffer.limit(writePosition+remaining);		
		buffer.put(source);
		unread+=remaining;
		source.limit(oldLimit);
		
	//	System.out.println("after: readPosition: "+readPosition+", unread: "+unread);
	}
	
	private static int getObjectSize(ByteBuffer buffer) {
		return buffer.getShort(buffer.position()+2)&0xffff;
	}

	public void getNext(ByteBuffer dest) {
		//System.out.println("before: readPosition: "+readPosition+", unread: "+unread);
		reset();
		
		int size = nextObjectSize();

		if (dest.remaining() < size) {
			throw new RuntimeException(
					"destination does not have enough buffer space to write an object with length "
							+ size);
		}

		if (unread < size) {
			throw new RuntimeException(
					"buffer contains less data than object length");
		}

		
		
		if (readPosition+size > buffer.capacity()) {
			buffer.position(readPosition);
			buffer.limit(buffer.capacity());
			int remaining = buffer.remaining();
			readPosition=0;
			unread-=remaining;
			size-=remaining;
			dest.put(buffer);			
		}
		
		buffer.position(readPosition);
		buffer.limit(readPosition+size);
		dest.put(buffer);
		readPosition=(readPosition+size)%buffer.capacity();
		unread-=size;
		
		// System.out.println("after: readPosition: "+readPosition+", unread: "+unread);
		
	}

	private void removeNextObject() {
		if (isEmpty()) {
			throw new RuntimeException(
					"cannot remove object from buffer. the bufer is empty");
		}

		int size = nextObjectSize();
		readPosition=(readPosition+size) % buffer.capacity();
		unread-=size;
	}

	public boolean isEmpty() {
		return unread==0;
	}

	private final int nextObjectSize() {
		if (unread < 4) {
			throw new RuntimeException("buffer is empty");
		}
		int ret;
		
		if (readPosition == buffer.capacity()-3) {
			int hi = (buffer.get(readPosition+2) & 0xff);
			int lo = (buffer.get(0) & 0xff);
			ret = (hi << 8) + lo;
		} else if (readPosition == buffer.capacity()-2){
			ret = buffer.getShort(0) & 0xffff;
		} else if (readPosition == buffer.capacity()-1) {
			ret = buffer.getShort(1) & 0xffff;;
		} else {
		
			ret = buffer.getShort(readPosition+2) & 0xffff;
		}
		

		return ret;

	}

	private final int remainingBytes() {
		return buffer.capacity()-unread;
	}

	private final void reset() {
		this.buffer.position(0);
		this.buffer.limit(buffer.capacity());
	}

	public final boolean isFull() {
		return buffer.capacity() == unread;
	}
	
	
	public static void  main(String[] argv) {
		ByteBuffer b9=ByteBuffer.allocate(100);
		ByteBuffer b10=ByteBuffer.allocate(100);
		ByteBuffer b11=ByteBuffer.allocate(100);
		ByteBuffer b2=ByteBuffer.allocate(100);
		ByteBuffer c=ByteBuffer.allocate(40);
		CircularMessageBuffer cmb=new CircularMessageBuffer(c);
		createMessage(b9,9);
		createMessage(b10,10);
		createMessage(b11,11);
		
		b10.flip();
		cmb.add(b10);
		
		b10.flip();
		cmb.add(b10);
		
		b10.flip();
		cmb.add(b10);
		
		b9.flip();
		cmb.add(b9);

		b11.flip();
		cmb.add(b11);
		
		
		b2.clear();
		cmb.getNext(b2);
		checkMessage(b2,10);
		
		b2.clear();
		cmb.getNext(b2);
		checkMessage(b2,10);
		
		b2.clear();
		cmb.getNext(b2);
		checkMessage(b2,9);
		
		b2.clear();
		cmb.getNext(b2);
		checkMessage(b2,11);
		
		if (!cmb.isEmpty()) {
			throw new RuntimeException("not empty!");			
		}
		
		
	}

	private static void checkMessage(ByteBuffer b2, int s) {
		b2.flip();
		int pad=b2.getShort();
		int size=b2.getShort()&0xffff;
		if (size != s) {
			throw new RuntimeException("failed "+s+" != "+size);
		}
		for (int i=0; i<size-4; i++) {
			if (b2.get() != (byte)i) {
				throw new RuntimeException("failed");
			}
				
		}
		
		
	}

	private static void createMessage(ByteBuffer b1, int size) {
		b1.clear();	
		b1.putShort((short)0);
		b1.putShort((short)size);
		for (int i=0; i<(size-4); i++) {
			b1.put((byte)i);
		}
	}

}
