package ch.unibe.iam.mcast.core.java;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedByInterruptException;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.unibe.iam.mcast.core.ConnectCallback;
import ch.unibe.iam.mcast.core.ConnectionStatusEventListener;
import ch.unibe.iam.mcast.core.LocalNetworkInterface;
import ch.unibe.iam.mcast.core.LocalNetworkInterfaceEventListener;
import ch.unibe.iam.mcast.core.MulticastCore;
import ch.unibe.iam.mcast.core.MulticastRoutingDestination;
import ch.unibe.iam.mcast.core.MulticastRoutingTable;
import ch.unibe.iam.mcast.core.MulticastRoutingTableEntry;
import ch.unibe.iam.mcast.core.MulticastRoutingTableListener;
import ch.unibe.iam.mcast.core.PeerConnection;
import ch.unibe.iam.mcast.core.PeerConnectionListener;
import ch.unibe.iam.mcast.net.ip.IPProtocolConstants;
import ch.unibe.iam.mcast.router.p2p.message.P2PMessage;

public class JavaMulticastCore implements MulticastCore, PeerConnectionListener, LocalNetworkInterfaceEventListener {

	private static final int OPEN_QUEUE_SIZE = 30;

	private static final Log log = LogFactory.getLog(JavaMulticastCore.class);

	private final InetSocketAddress address;

	private final JavaMulticastRoutingTable table = new JavaMulticastRoutingTable();

	private final HashSet<JavaPeerConnection> openConnections = new HashSet<JavaPeerConnection>();

	private final HashSet<JavaTapNetworkInterface> openInterfaces = new HashSet<JavaTapNetworkInterface>();

	private final ArrayList<ConnectionStatusEventListener> connectionStatusEventListeners = new ArrayList<ConnectionStatusEventListener>();

	private final java.util.concurrent.LinkedBlockingQueue<RemoteAddressCallbackPair> openQueue = new java.util.concurrent.LinkedBlockingQueue<RemoteAddressCallbackPair>(OPEN_QUEUE_SIZE);

	private ConnectingThread connectingThread;

	private ServerThread serverThread;

	private final DataForwarder forwarder = new DataForwarder();

	public JavaMulticastCore(InetSocketAddress address) {
		if (address == null) {
			throw new IllegalArgumentException("address should not be null");
		}
		this.address = address;
	}

	public MulticastRoutingTable getRoutingTable() {
		return table;
	}

	public List<PeerConnection> getCurrentPeerConnections() {
		return new ArrayList<PeerConnection>(openConnections);
	}

	public synchronized LocalNetworkInterface openVirtualInterface(InetAddress interfaceAddress) throws IOException {
		JavaTapNetworkInterface iface = new JavaTapNetworkInterface(interfaceAddress);
		iface.addLocalNetworkInterfaceEventListener(this);
		openInterfaces.add(iface);
		iface.addRawIPPacketListener(forwarder);
		// iface.addIPV4PacketListener(IPProtocolConstants.TCP, forwarder);

		iface.start();
		return iface;
	}

	public void onPeerMessage(PeerConnection connection, P2PMessage message) {
		// not interested in messages

	}

	public synchronized void onPeerConnectionClosed(PeerConnection connection) {
		openConnections.remove(connection);
	}

	public void connectToPeer(InetSocketAddress remotePeer, ConnectCallback callback) {
//		System.out.println("putting open request for "+remotePeer+" to connect queue");
		try {			
			openQueue.put(new RemoteAddressCallbackPair(remotePeer, callback));
		} catch (InterruptedException e) {
			log.error(e);
			Thread.currentThread().interrupt();
		}
//		System.out.println("put done");

	}

	public void addConnectionStatusEventListener(ConnectionStatusEventListener listener) {
		if (!connectionStatusEventListeners.contains(listener)) {
			connectionStatusEventListeners.add(listener);
		}

	}

	public void removeConnectionStatusEventListener(ConnectionStatusEventListener listener) {
		connectionStatusEventListeners.remove(listener);
	}

	private void onConnectFailed(InetSocketAddress peer, IOException reason) {
		for (int i = 0; i < connectionStatusEventListeners.size(); i++) {
			ConnectionStatusEventListener listener = (ConnectionStatusEventListener) connectionStatusEventListeners.get(i);
			listener.onPeerConnectFailed(peer, reason);
		}

	}

	private void onNewConnection(JavaPeerConnection peer) {
		for (int i = 0; i < connectionStatusEventListeners.size(); i++) {
			ConnectionStatusEventListener listener = (ConnectionStatusEventListener) connectionStatusEventListeners.get(i);
			listener.onNewPeerConnectionOpen(peer);
		}

	}

	public synchronized void start() throws IOException {
		if (connectingThread != null) {
			throw new IllegalStateException("already running");
		}

		this.serverThread = new ServerThread();
		this.serverThread.start();

		this.connectingThread = new ConnectingThread();
		this.connectingThread.start();

	}

	public void stop() {

		HashSet<JavaPeerConnection> open = new HashSet<JavaPeerConnection>();

		synchronized (this) {
			if (connectingThread == null) {
				throw new IllegalStateException("not running");
			}
			serverThread.interrupt();
			try {
				serverThread.join();
			} catch (InterruptedException e) {
				// just ignore
			}

			connectingThread.interrupt();
			try {
				connectingThread.join();
			} catch (InterruptedException e) {
				// just ignore
			}

			// TODO close all connections

			connectingThread = null;

			open = new HashSet<JavaPeerConnection>(openConnections);
		}

		
		
		
		for (Iterator<JavaPeerConnection> i = open.iterator(); i.hasNext();) {
			JavaPeerConnection conn = (JavaPeerConnection) i.next();
			conn.close();
		}
	}

	private class JavaMulticastRoutingTable implements MulticastRoutingTable {

		private final ArrayList<MulticastRoutingTableListener> listeners = new ArrayList<MulticastRoutingTableListener>();

		private final Map<InetAddress, ArrayList<MulticastRoutingTableEntry>> destinations = new HashMap<InetAddress, ArrayList<MulticastRoutingTableEntry>>(); // InetAddress

		// ->

		public String toString() {
			StringBuffer ret = new StringBuffer();
			ret.append("\nMulticastRoutingTable:\n");
			ArrayList<InetAddress> keys = new ArrayList<InetAddress>(destinations.keySet());
			// Comparator <InetAddress> comparator=new Comparator<InetAddress> {
			// }
			// }
			Collections.sort(keys, new Comparator<InetAddress>() {
				public int compare(InetAddress o1, InetAddress o2) {
					byte[] a1 = o1.getAddress();
					byte[] a2 = o2.getAddress();
					if (a1.length != a2.length) {
						return a1.length - a2.length;
					}
					for (int i = 0; i < a1.length; i++) {
						if (a1[i] != a2[i]) {
							return (a1[i] & 0xff) - (a2[i] & 0xff);
						}
					}
					return 0;
				}
			});
			for (InetAddress group : keys) {
				ret.append(">");
				ret.append(group.getHostAddress());
				ret.append(": \n");
				for (MulticastRoutingTableEntry e : destinations.get(group)) {
					ret.append("  ");
					ret.append(e);
					ret.append("\n");
				}

			}
			return ret.toString();
		}

		// Set(MulticastRoutingDestination)

		// private final Map destinationPriorities = new HashMap(); //
		// InetAddress -> Integer (priority)

		public synchronized void addRoute(InetAddress multicastGroup, MulticastRoutingDestination destination, int priority) {
			System.out.println("+++ adding route " + multicastGroup + " -> " + destination);

			ArrayList<MulticastRoutingTableEntry> destinationList = destinations.get(multicastGroup);
			if (destinationList == null) {
				destinationList = new ArrayList<MulticastRoutingTableEntry>();
				destinations.put(multicastGroup, destinationList);
			}

			boolean exists = false;
			MulticastRoutingTableEntry newEntry = new MulticastRoutingTableEntry(destination, priority);
			for (Iterator<MulticastRoutingTableEntry> i = destinationList.iterator(); i.hasNext();) {
				MulticastRoutingTableEntry entry = i.next();
				if (entry.getDestination().equals(destination)) {
					if (entry.getPriority() != priority) {
						i.remove();
					} else {
						exists = true;
					}
				}
			}

			System.out.println("   exists: "+exists);
			if (!exists) {
				destinationList.add(newEntry);
				onRouteAdded(multicastGroup, newEntry);
			}

		}

		private synchronized void onRouteAdded(InetAddress multicastGroup, MulticastRoutingTableEntry entry) {
			for (int i = 0; i < listeners.size(); i++) {
				((MulticastRoutingTableListener) listeners.get(i)).multicastRouteAdded(this, multicastGroup, entry);
			}

		}

		private synchronized void onRouteRemoved(InetAddress multicastGroup, MulticastRoutingTableEntry entry) {
			for (int i = 0; i < listeners.size(); i++) {
				((MulticastRoutingTableListener) listeners.get(i)).multicastRouteRemoved(this, multicastGroup, entry);
			}

		}

		public synchronized Map<InetAddress,Set<MulticastRoutingTableEntry>> getRoutes() {
			HashMap<InetAddress,Set<MulticastRoutingTableEntry>> ret = new HashMap<InetAddress,Set<MulticastRoutingTableEntry>>();
			for (Iterator<Map.Entry<InetAddress, ArrayList<MulticastRoutingTableEntry>>> i = destinations.entrySet().iterator(); i.hasNext();) {
				Map.Entry<InetAddress, ArrayList<MulticastRoutingTableEntry>> entry = i.next();
				InetAddress multicastGroup = entry.getKey();
				ArrayList<MulticastRoutingTableEntry> values = entry.getValue();
				ret.put(multicastGroup, new HashSet<MulticastRoutingTableEntry>(values));
			}
			return ret;
		}

		public synchronized Set<MulticastRoutingTableEntry> getDestinations(Inet4Address multicastGroup) {
			ArrayList<MulticastRoutingTableEntry> ret = destinations.get(multicastGroup);

			return ret == null ? Collections.<MulticastRoutingTableEntry>emptySet(): new HashSet<MulticastRoutingTableEntry>(ret);
		}
		

		private synchronized ArrayList<MulticastRoutingTableEntry> getDestinationList(Inet4Address multicastGroup) {
			return destinations.get(multicastGroup);
		}

		public synchronized void removeRoute(InetAddress multicastGroup, MulticastRoutingDestination destination) {
			ArrayList<MulticastRoutingTableEntry> destinationList = destinations.get(multicastGroup);
			if (destinationList != null) {

				removeDestinationFromSet(multicastGroup, destination, destinationList);

			} else {
				log.warn("trying to remove not-existing route " + multicastGroup + " -> " + destination, new RuntimeException());

			}

		}

		/**
		 * @param multicastGroup
		 * @param destination
		 * @param destinationList
		 */
		private void removeDestinationFromSet(InetAddress multicastGroup, MulticastRoutingDestination destination, ArrayList<MulticastRoutingTableEntry> destinationList) {
			MulticastRoutingTableEntry removedEntry = null;

			for (Iterator<MulticastRoutingTableEntry> i = destinationList.iterator(); i.hasNext();) {
				MulticastRoutingTableEntry entry = (MulticastRoutingTableEntry) i.next();
				if (entry.getDestination().equals(destination)) {
					removedEntry = entry;
					i.remove();

				}

			}
			if (removedEntry != null) {
				onRouteRemoved(multicastGroup, removedEntry);
			} else {
				// log.warn("trying to remove not-existing route " +
				// multicastGroup + " -> " + destination, new
				// RuntimeException());
			}
		}

		public synchronized void removeDestination(MulticastRoutingDestination destinationToRemove) {
			for (Iterator<Map.Entry<InetAddress, ArrayList<MulticastRoutingTableEntry>>> i = destinations.entrySet().iterator(); i.hasNext();) {

				Map.Entry<InetAddress, ArrayList<MulticastRoutingTableEntry>> entry =  i.next();
				InetAddress group = entry.getKey();
				ArrayList<MulticastRoutingTableEntry> dsts = entry.getValue();

				removeDestinationFromSet(group, destinationToRemove, dsts);

				if (dsts.isEmpty()) {
					i.remove();
				}
			}
		}

		public synchronized boolean hasRoutesForGroup(InetAddress multicastGroup) {
			ArrayList<MulticastRoutingTableEntry> dsts =  destinations.get(multicastGroup);
			if (dsts == null) {
				return false;
			}
			return !dsts.isEmpty();
		}

		public synchronized void addMulticastRoutingTableListener(MulticastRoutingTableListener listener) {
			if (!listeners.contains(listener)) {
				listeners.add(listener);
			}
		}

		public synchronized void removeMulticastRoutingTableListener(MulticastRoutingTableListener listener) {
			listeners.remove(listener);
		}

		public synchronized boolean hasRoute(InetAddress group, MulticastRoutingDestination destination) {
			if (!destinations.containsKey(group)) {
				return false;
			} else {
				return ( destinations.get(group)).contains(destination);
			}
		}

		public synchronized int getMaxPriority(InetAddress group) {
			int min = Integer.MAX_VALUE;

			ArrayList<MulticastRoutingTableEntry> destinationPriorityList = this.destinations.get(group);
			if (destinationPriorityList != null) {

				for (int i = 0; i < destinationPriorityList.size(); i++) {
					MulticastRoutingTableEntry entry = destinationPriorityList.get(i);
					min = Math.min(min, entry.getPriority());
				}
			} else {
				log.warn("trying to determine maximum for an empty group " + group);
			}

			return min;

		}

		public int getMaxPriority(InetAddress group, MulticastRoutingDestination destination) {
			int min = MulticastRoutingTable.PRIORITY_LOW;

			ArrayList<MulticastRoutingTableEntry> destinationSet = this.destinations.get(group);
			boolean single = destinationSet.size() <= 1;

			if (destinationSet != null) {
				for (Iterator<MulticastRoutingTableEntry> i = destinationSet.iterator(); i.hasNext();) {

					MulticastRoutingTableEntry entry = (MulticastRoutingTableEntry) i.next();
					if (single || !entry.getDestination().equals(destination)) {
						min = Math.min(min, entry.getPriority());
					}
				}
			} else {
				log.warn("trying to determine maximum for an empty group " + group);
			}

			return min;
		}

	}

	private void registerNewConnection(JavaPeerConnection newConn) {

		newConn.addPeerMessageListener(this);
		openConnections.add(newConn);
		onNewConnection(newConn);
		newConn.addDataMessageListener(forwarder);
		newConn.start();
	}

	private class ConnectingThread extends Thread {

		public ConnectingThread() {
			super("Connecting Thread");
			setDaemon(true);
		}

		public void run() {

			while (!interrupted()) {
				InetSocketAddress peerAddress = null; // just to satisfy the
				// try{} ...
				RemoteAddressCallbackPair pair;
				try {
					pair = openQueue.take();
				} catch (InterruptedException e1) {
					log.error(e1);
					return;
				}
//				System.out.println("opening connection to "+pair.getRemoteAddress().getAddress()+"/"+pair.getRemoteAddress().getPort());
				try {
					SocketChannel channel = SocketChannel.open(pair.getRemoteAddress());
					JavaPeerConnection newConn = new JavaPeerConnection(channel);
					registerNewConnection(newConn);
					pair.getCallback().connectionOpen(newConn);
				} catch (IOException e) {
					if (log.isDebugEnabled())
						log.debug(e);
					onConnectFailed(peerAddress, e);
					pair.getCallback().connectFailed(pair.getRemoteAddress(), e);
				} catch (RuntimeException re) {
					re.printStackTrace();
//					log.error(re);
				}
			}

		}

		/**
		 * @param newConn
		 */

	}

	private class ServerThread extends Thread {

		private final ServerSocketChannel serverChannel;

		public ServerThread() throws IOException {
			super("Accepting P2P Connections");
			setDaemon(true);			
			this.serverChannel = ServerSocketChannel.open();
			serverChannel.socket().bind(address);

		}

		public void run() {

			try {
				while (!interrupted()) {
					SocketChannel newSocket = serverChannel.accept();
					registerNewConnection(new JavaPeerConnection(newSocket));
				}
			} catch (ClosedByInterruptException e) {
				// just close
			} catch (IOException e) {
				log.warn(e);
			} finally {
				try {
					serverChannel.close();
				} catch (IOException e) {
					log.warn(e);
				}

			}

		}
	}

	public class DataForwarder implements DataMessageListener, RawIPPacketListener {

		private IPV4AddressCache addressCache = new IPV4AddressCache();

		// public void onPeerMessage(PeerConnection connection, P2PMessage
		// message) {
		// if (message instanceof IPV4DataPacketMessage) {
		// IPV4DataPacketMessage msg = (IPV4DataPacketMessage) message;
		// IPV4Packet packet = msg.getPacket();
		// doForward(packet, connection);
		// }
		//
		// }

		/**
		 * @param packet
		 */
		// private void doForward(IPV4Packet packet, MulticastRoutingDestination
		// source) {
		// Inet4Address destinationAddress = packet.getDestinationAddress();
		// Set destinations = table.getDestinations(destinationAddress);
		//
		// if (!destinations.isEmpty()) {
		// int priority = table.getMaxPriority(destinationAddress);
		// for (Iterator i = destinations.iterator(); i.hasNext();) {
		// MulticastRoutingTableEntry entry = (MulticastRoutingTableEntry)
		// i.next();
		// MulticastRoutingDestination destination = entry.getDestination();
		// if (!source.equals(destination)) {
		// destination.sendIPV4Packet(packet, priority);
		// }
		// }
		// }
		// }
		private void doForwardRaw(ByteBuffer packet, MulticastRoutingDestination source) {
			// System.out.println("received buffer: "+packet);
			// int orig=packet.position();
			// packet.position(0);
			// ByteBufferUtil.dumpByteBuffer(packet);
			// packet.position(orig);
			if (packet.remaining() >= IPProtocolConstants.IPV4_DESTINATION_ADDRESS_OFFSET + 4) {

				int ttl = packet.get(packet.position() + IPProtocolConstants.IPV4_TTL_OFFSET) & 0xff;
				if (ttl < 2) {
					return;
				}

				int dstAddr = packet.getInt(packet.position() + IPProtocolConstants.IPV4_DESTINATION_ADDRESS_OFFSET);
				Inet4Address destinationAddress = addressCache.lookup(dstAddr);
				ArrayList<MulticastRoutingTableEntry> destinations = table.getDestinationList(destinationAddress);

				// System.out.println("forwarding "+destinationAddress+" from
				// "+source+" to "+destinations);
				int origPosition = packet.position();
				int origLimit = packet.limit();
				if (destinations != null && !destinations.isEmpty()) {

					// int priority = table.getMaxPriority(destinationAddress);
					for (int i = 0; i < destinations.size(); i++) {
						MulticastRoutingTableEntry entry = destinations.get(i);
						int priority = entry.getPriority();
						JavaMulticastRoutingDestination destination = (JavaMulticastRoutingDestination) entry.getDestination();
						if (!source.equals(destination)) {
							packet.position(origPosition);
							packet.limit(origLimit);
							destination.sendRawIPV4Packet(packet, priority);
						}
					}
				}
			}
		}

		public void onPeerConnectionClosed(PeerConnection connection) {
			// TODO clean up the multicast table

		}

		// public void onIPPacket(LocalNetworkInterface iface, IPV4Packet
		// packet) {
		//            
		//
		// }

		public void onNewRawIPPacket(LocalNetworkInterface iface, ByteBuffer packet) {
			doForwardRaw(packet, iface);

		}

		public void onNewDataMessage(JavaPeerConnection source, ByteBuffer message) {
			// System.out.println("got peer data");
			doForwardRaw(message, source);

		}

	}

	public synchronized void onLocalNetworkInterfaceShutdown(LocalNetworkInterface iface) {
		openInterfaces.remove(iface);

	}

	// public static void main(String argv[]) throws Exception {
	// JavaMulticastCore jmc = new JavaMulticastCore(new
	// InetSocketAddress("localhost", 6666));
	// jmc.start();
	// jmc.addConnectionStatusEventListener(new TestClass());
	// jmc.connectToPeer(new InetSocketAddress("localhost", 6666));
	// Thread.sleep(200);
	// jmc.stop();
	//
	// }

	// private static class TestClass implements ConnectionStatusEventListener,
	// PeerConnectionListener {
	//
	// public void onNewPeerConnectionOpen(PeerConnection connection) {
	// System.out.println("connected to " + connection);
	// connection.addPeerMessageListener(this);
	// connection.sendMessage(new KeepAliveMessage());
	// }
	//
	// public void onPeerConnectFailed(InetSocketAddress connection, IOException
	// reason) {
	// System.out.println("failed to connect to " + connection);
	// reason.printStackTrace();
	//
	// }
	//
	// public void onPeerMessage(PeerConnection connection, P2PMessage message)
	// {
	// System.out.println("received " + message + " message from " +
	// connection);
	//
	// }
	//
	// public void onPeerConnectionClosed(PeerConnection connection) {
	// System.out.println("connection to " + connection + " closed");
	//
	// }
	//
	// }

	private static class RemoteAddressCallbackPair {
		private final InetSocketAddress remoteAddress;

		private final ConnectCallback callback;

		public RemoteAddressCallbackPair(InetSocketAddress remoteAddress, ConnectCallback callback) {
			this.remoteAddress = remoteAddress;
			this.callback = callback;

		}

		public ConnectCallback getCallback() {
			return callback;
		}

		public InetSocketAddress getRemoteAddress() {
			return remoteAddress;
		}

	}

	public int getServerPort() {
		return address.getPort();
	}
}
