package ch.unibe.iam.mcast.core.java;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.unibe.iam.mcast.core.IPV4PacketListener;
import ch.unibe.iam.mcast.core.LocalNetworkInterface;
import ch.unibe.iam.mcast.core.LocalNetworkInterfaceEventListener;
import ch.unibe.iam.mcast.iface.linux.tuntap.TapInterface;
import ch.unibe.iam.mcast.net.Packet;
import ch.unibe.iam.mcast.net.ethernet.EthernetAddress;
import ch.unibe.iam.mcast.net.ethernet.EthernetConstants;
import ch.unibe.iam.mcast.net.ethernet.EthernetPacket;
import ch.unibe.iam.mcast.net.ip.IPProtocolConstants;
import ch.unibe.iam.mcast.net.ip.IPV4Packet;
import ch.unibe.iam.mcast.net.ip.MalformedPacketException;

public class JavaTapNetworkInterface extends LocalNetworkInterface implements JavaMulticastRoutingDestination {

    private static final int MAX_PACKET_SIZE = 4096;

    private static final Log log = LogFactory.getLog(JavaTapNetworkInterface.class);

    private final TapInterface realInterface;

    private final TapThread readingThread = new TapThread();

    private final EthernetAddress interfaceAddress = new EthernetAddress(new byte[] { 0x00,0x0B,0x6B,0x57,0x3E,0x9 });
    
     
    
    private final byte[] bufferArray = new byte[MAX_PACKET_SIZE];

    private final ByteBuffer buffer = ByteBuffer.wrap(bufferArray);
    
    private final List rawIPPacketListeners=new ArrayList();

    public JavaTapNetworkInterface(InetAddress address) throws IOException {
        super("tap0", address);
        this.realInterface = TapInterface.openNewTapInterface((Inet4Address) address);
    }

    public void start() {
        this.readingThread.start();
    }

    public synchronized void sendIPV4Packet(IPV4Packet packet, int priority) {
        
        // FIXME priorities???
        
        EthernetAddress destinationAddress;
        if (packet.getDestinationAddress().isMulticastAddress()) {
            destinationAddress = createMulticastAddress(packet.getDestinationAddress());
        } else {
            destinationAddress = interfaceAddress;
        }

        EthernetPacket ep = new EthernetPacket(destinationAddress, interfaceAddress, packet);

        buffer.clear();
        int packetLength = ep.getPacketLength();
        buffer.limit(packetLength);
        ep.serializePacket(buffer);
        try {
            realInterface.writePacket(bufferArray, packetLength);
        } catch (IOException e) {
            log.error(e);
        }
    }

    private static EthernetAddress createMulticastAddress(Inet4Address address) {
        // 01:00:5e
        byte[] inet = address.getAddress();
        byte third = (byte) (inet[1] & 0x7f);
        return new EthernetAddress(new byte[] { 0x01, 0x00, 0x5e, third, inet[2], inet[3] });
    }
    
    private static void serializeMulticastAddress(ByteBuffer buffer, int address) {
//    	System.out.println("-> serializing "+address);
    	buffer.put((byte) 0x01);
    	buffer.put((byte) 0x00);
    	int mangled=(0x5e<<24)|(address & 0x007fffff);
//    	System.out.println("-> mangled "+mangled);
    	buffer.putInt(mangled);
    }
    

    private class TapThread extends Thread {
        private static final int IPV4_PROTOCOL_NUMBER_OFFSET = 9;

		private static final int ETHERNET_HEADER_LENGTH = 14;

		private final byte[] buffer = new byte[8192];

        private final ByteBuffer wrapper = ByteBuffer.wrap(buffer);

        public void run() {
            try {
                while (!interrupted()) {
                    int read = realInterface.readPacket(buffer);
                    wrapper.clear();
                    wrapper.limit(read);
                    
                    if (PacketHelper.isIPV4Packet(wrapper)) {
                    	
                    	byte proto=wrapper.get(wrapper.position()+ETHERNET_HEADER_LENGTH+IPV4_PROTOCOL_NUMBER_OFFSET);
                    	
                    	int originalPosition=wrapper.position();
                    	int originalLimit=wrapper.limit();
                    	
                    	
                    	if (hasCookedListener(proto)) {
                        	try {
                                EthernetPacket ep = EthernetPacket.createPacket(wrapper);

                                Packet transported = ep.getTransportedPacket();
                                if (transported instanceof IPV4Packet) {
                                    onIPV4Packet((IPV4Packet) transported);
                                }

                            } catch (MalformedPacketException e) {
                                log.warn(e);
                            }
                    	}
                    	
                    	
                    	for (int i=0; i<rawIPPacketListeners.size(); i++) {
                    		RawIPPacketListener listener=(RawIPPacketListener) rawIPPacketListeners.get(i);
                    		wrapper.position(originalPosition+ETHERNET_HEADER_LENGTH);
                    		wrapper.limit(originalLimit);
                    		listener.onNewRawIPPacket(JavaTapNetworkInterface.this,wrapper); 
                    	}
                    	
                    	
                    } else {
                    	// what should be done with non-ethernet packets?
                    	
                    }
                    
                    
                }
            } catch (IOException e) {
                log.warn(e);
            } finally {
                try {
                    realInterface.close();
                } catch (IOException e) {
                    log.warn(e);
                }
            }

        }
    }

    public void close() {
        this.readingThread.interrupt();
//        try {
//            realInterface.close();
//        } catch (IOException e) {
//            log.error(e);
//        }
    }

    public static void main(String argv[]) throws Exception {
        JavaTapNetworkInterface jtni = new JavaTapNetworkInterface(Inet4Address.getByAddress(new byte[] { 1, 2, 3, 4 }));
        jtni.addLocalNetworkInterfaceEventListener(new LocalNetworkInterfaceEventListener() {
            public void onLocalNetworkInterfaceShutdown(LocalNetworkInterface iface) {
                System.out.println(iface.getName() + " shut down");

            }
        });

        jtni.addIPV4PacketListener(IPProtocolConstants.UDP, new IPV4PacketListener() {

            public void onIPPacket(LocalNetworkInterface iface, IPV4Packet packet) {
                System.out.println("packet received " + packet);

            }

        });
        
        jtni.addIPV4PacketListener(IPProtocolConstants.ICMP, new IPV4PacketListener() {

            public void onIPPacket(LocalNetworkInterface iface, IPV4Packet packet) {
                System.out.println("ICMP packet received " + packet);

            }

        });
        
        jtni.start();
        
        Thread.sleep(112000);
        
        jtni.close();
        
        
    }

    public String toString() {
        return "TapInterface("+super.getName()+")";
    }
    
    
    public synchronized void addRawIPPacketListener(RawIPPacketListener listener) {
    	this.rawIPPacketListeners.add(listener);
    }

    public synchronized void removeRawIPPacketListener(RawIPPacketListener listener) {
    	this.rawIPPacketListeners.remove(listener);
    }

	public synchronized void sendRawIPV4Packet(ByteBuffer packet, int priority) {
		
		 buffer.clear();
		 
		 
		 
		
		 //EthernetAddress destinationAddress;
		 
		 int destinationAddress=packet.getInt(packet.position()+IPProtocolConstants.IPV4_DESTINATION_ADDRESS_OFFSET);
		 
		 if ((destinationAddress & 0xf0000000) == 0xe0000000) {
			// multicast address
			serializeMulticastAddress(buffer,destinationAddress);
		 } else {
			interfaceAddress.serializeAddress(buffer);
		 }
		 
		 interfaceAddress.serializeAddress(buffer);
		 
		 buffer.putShort(EthernetConstants.ETH_P_IP);
		 buffer.put(packet);
		 
		 
	        try {
	            realInterface.writePacket(bufferArray, buffer.position());
	        } catch (IOException e) {
	            log.error(e);
	        }	
		
	}

    

}
