/*
 * Created on Oct 7, 2004
 *
 */
package ch.unibe.iam.mcast.net.ip;

import java.net.Inet4Address;
import java.nio.ByteBuffer;
import java.util.List;

import ch.unibe.iam.mcast.net.Packet;
import ch.unibe.iam.mcast.net.ip.icmp.ICMPV4PacketFactory;
import ch.unibe.iam.mcast.net.ip.igmp.IGMPProtocolPacketFactory;
import ch.unibe.iam.mcast.net.ip.udp.UDPPacketFactory;
import ch.unibe.iam.mcast.net.ip.util.InternetChecksumUtil;

/**
 * @author dragan
 *  
 */
public abstract class IPV4Packet implements Packet {

    private static final int PROTOCOL_OFFSET = 9;

    private static final IPProtocolPacketFactory[] packetFactories = new IPProtocolPacketFactory[256];

    static {
        UnknownIPPacketProtocolFactory unknown = new UnknownIPPacketProtocolFactory();
        for (int i = 0; i < 255; i++) {
            packetFactories[i] = unknown;
        }

        packetFactories[IPProtocolConstants.IGMP & 0xff]=new IGMPProtocolPacketFactory();
        packetFactories[IPProtocolConstants.ICMP & 0xff]=new ICMPV4PacketFactory();
        packetFactories[IPProtocolConstants.UDP & 0xff]=new UDPPacketFactory();
    }

    public synchronized static IPV4Packet createIPV4Packet(ByteBuffer buffer) throws MalformedPacketException {
        byte protocol = buffer.get(buffer.position() + PROTOCOL_OFFSET);
        
        //System.out.println("protocol: "+protocol);

        IPV4PacketHeader header = new IPV4PacketHeader(buffer);

        IPProtocolPacketFactory factory = packetFactories[protocol & 0xff];
        IPV4Packet packet;

        packet = factory.createPacket(header, buffer);

        return packet;

    }

    private static final int TOTAL_LENGTH_OFFSET = 2;

    private static final int CHECKSUM_OFFSET = 10;

    protected final IPV4PacketHeader header;

    public IPV4Packet(byte protocol) {
        this.header = new IPV4PacketHeader();
        header.setProtocol(protocol);

    }

    public IPV4Packet(IPV4PacketHeader header) {
        this.header = header;

    }

    

    public void serializePacket(ByteBuffer buffer) {
        int startPosition = buffer.position();
        int endPosition = buffer.limit();
        int checksumPosition = startPosition + CHECKSUM_OFFSET;
        header.serializeHeader(buffer);
        int headerEnd=buffer.position();        
        serializeData(buffer);
        int dataLength=buffer.position()-headerEnd;
        
        
        int totalPacketSize = getPacketLength();
        
        //System.out.println("total ip packet size: "+totalPacketSize);
        
        int expectedStart = (buffer.position() - totalPacketSize);
		if (expectedStart != startPosition) {
            throw new RuntimeException("ip packet implementation did not set the whole data buffer. start position: "+startPosition+" realStartPosition: "+expectedStart+" data length "+dataLength);
        }

        // calculate the IP-Checksum of the packet

        buffer.position(startPosition);
        buffer.limit(startPosition + totalPacketSize);


        buffer.putShort(buffer.position() + TOTAL_LENGTH_OFFSET, (short) totalPacketSize);

        
//        //buffer.position(checksumPosition);
        int oldLimit=buffer.limit();
        buffer.limit(headerEnd);
        short checksum = calculateIPChecksum(buffer);
        buffer.putShort(checksumPosition,checksum);
        buffer.limit(oldLimit);
        // reset the buffer pointers
        buffer.position(startPosition + totalPacketSize);
        buffer.limit(endPosition);

    }

    /**
     * @param buffer
     * @return
     */
    private static short calculateIPChecksum(ByteBuffer buffer) {
        return InternetChecksumUtil.computeInetChecksum(buffer);
    }

    public int getPacketLength() {
        return header.getHeaderLength() + getDataLengt();
    }

    public boolean equals(Object obj) {
        // FIXME find optimal method for equals
        if (!(obj instanceof IPV4Packet)) {
            return false;
        } else {
            throw new RuntimeException("not yet implemented");
        }
    }

    public int ahashCode() {
        throw new RuntimeException("not yet implemented");
    }

    protected abstract void serializeData(ByteBuffer buffer);

    protected abstract int getDataLengt();

    //protected abstract void deserializeData(ByteBuffer buffer);

    // DELEGATING GETTERS AND SETTERS BELOW

    public Inet4Address getDestinationAddress() {
        return header.getDestinationAddress();
    }

    public void setDestinationAddress(Inet4Address destinationAddress) {
        header.setDestinationAddress(destinationAddress);
    }

    public byte getFlags() {
        return header.getFlags();
    }

    public void setFlags(byte flags) {
        header.setFlags(flags);
    }

    public short getFragmentOffset() {
        return header.getFragmentOffset();
    }

    public void setFragmentOffset(short fragmentOffset) {
        header.setFragmentOffset(fragmentOffset);
    }

    public short getIdentification() {
        return header.getIdentification();
    }

    public void setIdentification(short identification) {
        header.setIdentification(identification);
    }

    public List getOptions() {
        return header.getOptions();
    }

    public void setOptions(List options) {
        header.setOptions(options);
    }

    public byte getProtocol() {
        return header.getProtocol();
    }

//    public void setProtocol(byte protocol) {
//        header.setProtocol(protocol);
//    }

    public Inet4Address getSourceAddress() {
        return header.getSourceAddress();
    }

    public void setSourceAddress(Inet4Address sourceAddress) {
        header.setSourceAddress(sourceAddress);
    }

    public byte getTimeToLive() {
        return header.getTimeToLive();
    }

    public void setTimeToLive(byte timeToLive) {
        header.setTimeToLive(timeToLive);
    }

    public byte getTypeOfService() {
        return header.getTypeOfService();
    }

    public void setTypeOfService(byte tos) {
        header.setTypeOfService(tos);
    }
    
    public boolean isDestinationMulticast() {
        return getDestinationAddress().isMulticastAddress();
    }

}