/*
 * Created on Oct 7, 2004
 *
 */
package ch.unibe.iam.mcast.net.ip;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import ch.unibe.iam.mcast.net.ip.option.HeaderOptionFactory;
import ch.unibe.iam.mcast.net.ip.util.ByteBufferUtil;

/**
 * @author dragan
 *  
 */
public class IPV4PacketHeader {

    private static final Inet4Address NULL_IPV4ADDRESS;

    static {
        byte[] address = new byte[4];
        try {
            NULL_IPV4ADDRESS = (Inet4Address) InetAddress.getByAddress(address);
        } catch (UnknownHostException e) {
            throw new RuntimeException("this should never happend", e);
        }
    }

    private byte typeOfService;
    
    private int totalLength;

    private short identification;

    private byte flags;

    private short fragmentOffset;

    private byte timeToLive;

    private byte protocol;

    private Inet4Address sourceAddress = NULL_IPV4ADDRESS;

    private Inet4Address destinationAddress = NULL_IPV4ADDRESS;

    private List options = new ArrayList();

    private static final int MIN_PACKET_SIZE = 5 * 4;

    public IPV4PacketHeader(ByteBuffer buffer) throws MalformedPacketException {

        deserializeHeader(buffer);

    }

    /**
     * @param buffer
     * @throws MalformedPacketException
     */
    public void deserializeHeader(ByteBuffer buffer) throws MalformedPacketException {
        int startPosition = buffer.position();

        int totalBufferSize = buffer.remaining();
        if (totalBufferSize < MIN_PACKET_SIZE) {
            throw new MalformedPacketException("packet size < " + MIN_PACKET_SIZE);
        }

        byte firstByte = buffer.get();
        int version = (firstByte & 0xf0) >> 4;
        
        
        if (version != 4) {
            throw new MalformedPacketException("not IPV4 Packet");
        }

        int headerLength = (firstByte & 0x0f) * 4;

        if (headerLength > totalBufferSize) {
            throw new MalformedPacketException("header length > total length");
        }

        this.typeOfService = buffer.get();

        totalLength = buffer.getShort() & 0xffff;
        

        if (totalLength > totalBufferSize) {
            throw new MalformedPacketException("total length beyond buffer size!");
        }

        this.identification = buffer.getShort();

        short flagsAndOffset = buffer.getShort();
        this.flags = (byte) ((flagsAndOffset & 0xe000) >> 13);
        this.fragmentOffset = (short) (flagsAndOffset & 0x1fff);
        this.timeToLive = buffer.get();
        this.protocol = buffer.get();
        short checksum = buffer.getShort();

        try {
            byte[] adrBuffer = new byte[4];
            buffer.get(adrBuffer);
            this.sourceAddress = (Inet4Address) InetAddress.getByAddress(adrBuffer);
            buffer.get(adrBuffer);
            this.destinationAddress = (Inet4Address) InetAddress.getByAddress(adrBuffer);
        } catch (UnknownHostException e) {
            throw new MalformedPacketException(e);
        }
        buffer.limit(startPosition+headerLength);
        parseOptions(buffer, totalLength, totalBufferSize);

        buffer.position(startPosition + headerLength);
        buffer.limit(startPosition + totalLength);
        

    }

    /**
     * @param buffer
     * @param totalLength
     * @param totalBufferSize
     */
    private void parseOptions(ByteBuffer buffer, int totalLength, int totalBufferSize) {
        while (buffer.remaining() > 0) {
            byte headerType = buffer.get();
            if (headerType != 0) {
                HeaderOption option = HeaderOptionFactory.createHeaderOption(headerType, buffer);
                options.add(option);
            }
        }
    }

    public void serializeHeader(ByteBuffer buffer) {
        int headerLength = getHeaderLength();
        int ihl = headerLength / 4;

        int startPosition = buffer.position();
        int endPosition = buffer.limit();
        buffer.put((byte) ((4 << 4) | ihl));
        buffer.put(typeOfService);
        
        int totalPacketSize = 0;
        buffer.putShort((short) totalPacketSize);
        buffer.putShort(identification);
        buffer.putShort((short) ((flags << 13) | (fragmentOffset & 0x1fff)));
        buffer.put(timeToLive);
        buffer.put(protocol);
        int checksumPosition = buffer.position();
        buffer.putShort((short) 0); // checksum
//        System.out.println("source address: "+sourceAddress);
//        ByteBufferUtil.dumpByteBuffer(ByteBuffer.wrap(sourceAddress.getAddress()));
        buffer.put(sourceAddress.getAddress());
        buffer.put(destinationAddress.getAddress());
        serializeOptions(buffer);
        
        
//        int oldLimit=buffer.limit();
//        int oldPoistion=buffer.position();
//        buffer.limit(buffer.position());
//        buffer.position(startPosition);
//        
//        short checksum=InternetChecksumUtil.computeInetChecksum(buffer);
//        
//        buffer.position(oldPoistion);
//        buffer.limit(oldLimit);
//        buffer.putShort(checksumPosition,checksum);
        

    }

    

    /**
     * @param buffer
     */
    private void serializeOptions(ByteBuffer buffer) {

        int startPosition = buffer.position();

        if (options.size() > 0) {
            for (Iterator i = options.iterator(); i.hasNext();) {
                HeaderOption option = (HeaderOption) i.next();
                option.serialize(buffer);
            }
            while ((buffer.position() - startPosition) % 4 != 0) {
                buffer.put((byte) 0);
            }
        }

    }

   

    /**
     * @return
     */
    public int getHeaderLength() {
        int ret = MIN_PACKET_SIZE;
        if (options.size() > 0) {
            for (Iterator i = options.iterator(); i.hasNext();) {
                HeaderOption option = (HeaderOption) i.next();
                ret += option.getLength();
            }
        }

        if (ret % 4 != 0) {
            // for the case that the options are not 4-byte alligned
            ret = ((ret / 4) + 1) * 4;
        }

        return ret;
    }

    public IPV4PacketHeader() {

    }

    public Inet4Address getDestinationAddress() {
        return destinationAddress;
    }

    public void setDestinationAddress(Inet4Address destinationAddress) {
        if (destinationAddress == null) {
            throw new IllegalArgumentException("destination address should not be null!");
        }
        this.destinationAddress = destinationAddress;
    }

    public byte getFlags() {
        return flags;
    }

    public void setFlags(byte flags) {
        this.flags = flags;
    }

    public short getFragmentOffset() {
        return fragmentOffset;
    }

    public void setFragmentOffset(short fragmentOffset) {
        this.fragmentOffset = fragmentOffset;
    }

    public short getIdentification() {
        return identification;
    }

    public void setIdentification(short identification) {
        this.identification = identification;
    }

    public List getOptions() {
        return options;
    }

    public void setOptions(List options) {
        this.options = options;
    }

    public byte getProtocol() {
        return protocol;
    }

    public void setProtocol(byte protocol) {
        this.protocol = protocol;
    }

    public Inet4Address getSourceAddress() {
        return sourceAddress;
    }

    public void setSourceAddress(Inet4Address sourceAddress) {
        if (sourceAddress == null) {
            throw new IllegalArgumentException("source address should not be null!");
        }
        this.sourceAddress = sourceAddress;
    }

    public byte getTimeToLive() {
        return timeToLive;
    }

    public void setTimeToLive(byte timeToLive) {
        this.timeToLive = timeToLive;
    }

    public boolean equals(Object obj) {
        if (!(obj instanceof IPV4PacketHeader)) {
            return false;
        } else {
            IPV4PacketHeader other = (IPV4PacketHeader) obj;
            return other.typeOfService == typeOfService
                   && other.identification == identification
                   && other.flags == this.flags
                   && other.fragmentOffset == this.fragmentOffset
                   && other.timeToLive == this.timeToLive
                   && other.protocol == this.protocol
                   && other.sourceAddress.equals(this.sourceAddress)
                   && other.destinationAddress.equals(this.destinationAddress)
                   && other.options.equals(this.options);

        }
    }

    public int hashCode() {
        return typeOfService
               ^ identification
               ^ flags
               ^ fragmentOffset
               ^ timeToLive
               ^ protocol
               ^ sourceAddress.hashCode()
               ^ destinationAddress.hashCode()
               ^ options.hashCode();
    }

    public byte getTypeOfService() {
        return typeOfService;
    }

    public void setTypeOfService(byte tos) {
        this.typeOfService = tos;
    }

    

    
    public int getTotalLength() {
        return totalLength;
    }
}