/*
 * Created on Nov 8, 2004
 *
 * Message format:
 *  0                   1                   2                   3
 *  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |     Type                      |    Total Length               |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |                                                               |
 * |                       Objects                                 |
 * |                                                               |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 * Type: 			type of the message (16 Bit)
 * Total Length: 	Total length of the message in bytes (including header)
 * Objects: 		List of the objects that compose the message
 */
package ch.unibe.iam.mcast.router.p2p.message;

import java.nio.ByteBuffer;

import ch.unibe.iam.mcast.router.p2p.message.data.IPV4DataPacketMessage;
import ch.unibe.iam.mcast.router.p2p.message.dynaddr.DynamicAddressResponseMessage;
import ch.unibe.iam.mcast.router.p2p.message.dynaddr.RequestDynamicAddressMessage;
import ch.unibe.iam.mcast.router.p2p.message.euqos.ChildHelloMessage;
import ch.unibe.iam.mcast.router.p2p.message.euqos.ParentHelloMessage;
import ch.unibe.iam.mcast.router.p2p.message.mcast.AddRouteMessage;
import ch.unibe.iam.mcast.router.p2p.message.mcast.OpenMessage;
import ch.unibe.iam.mcast.router.p2p.message.mcast.RemoveRouteMessage;

/**
 * @author root
 * 
 * TODO To change the template for this generated type comment go to Window - Preferences - Java - Code Style - Code Templates
 */
public abstract class P2PMessage {

    private static final Class[] messageClass = new Class[256];

    static {
        messageClass[P2PMessageConstants.OPEN_MESSAGE] = OpenMessage.class;
        messageClass[P2PMessageConstants.ADD_ROUTE_MESSAGE] = AddRouteMessage.class;
        messageClass[P2PMessageConstants.KEEP_ALIVE_MESSAGE] = KeepAliveMessage.class;
        messageClass[P2PMessageConstants.REMOVE_ROUTE_MESSAGE] = RemoveRouteMessage.class;
        
        messageClass[P2PMessageConstants.REQUEST_DYNAMIC_ADDRESS_MESSAGE] = RequestDynamicAddressMessage.class;
        messageClass[P2PMessageConstants.DYNAMIC_ADDRESS_RESPONSE_MESSAGE] = DynamicAddressResponseMessage.class;
       
        
        messageClass[P2PMessageConstants.PING_MESSAGE] = PingMessage.class;
        messageClass[P2PMessageConstants.PONG_MESSAGE] = PongMessage.class;
        messageClass[P2PMessageConstants.IPV4_DATA_PACKET_MESSAGE] = IPV4DataPacketMessage.class;
        messageClass[P2PMessageConstants.PARENT_HELLO_MESSAGE] = ParentHelloMessage.class;
        messageClass[P2PMessageConstants.CHILD_HELLO_MESSAGE] = ChildHelloMessage.class;
    }

    private final short type;

    /**
     *  
     */
    public P2PMessage(short type) {
        this.type = type;
    }

    private void deserialize(ByteBuffer buffer) throws MalformedP2PMessageException {
        if (buffer.remaining() < 2) {
            throw new MalformedP2PMessageException("message must be at least 2 bytes long");
        }
        int startLimit = buffer.limit();
        int startPosition = buffer.position();
        int length = buffer.getShort() & 0xffff;
        buffer.limit(startPosition + length - 2);
        deserializeContent(buffer);
        buffer.limit(startLimit);
    }

    public void serialize(ByteBuffer buffer) {
        buffer.putShort(type);
        int contentLength = contentLength();
        buffer.putShort((short) (contentLength + 4));
        int startLimit = buffer.limit();
        buffer.limit(buffer.position() + contentLength);
        serializeContent(buffer);
        buffer.limit(startLimit);

    }

    public int getMessageLength() {
        return contentLength() + 4;
    }

    protected abstract void serializeContent(ByteBuffer buffer);

    protected abstract int contentLength();

    protected abstract void deserializeContent(ByteBuffer buffer) throws MalformedP2PMessageException;

    public static P2PMessage deserializeMessage(ByteBuffer buffer) throws MalformedP2PMessageException {
        if (buffer.remaining() < 4) {
            throw new MalformedP2PMessageException("message must have at least 4 bytes");
        } else {
            int type = buffer.getShort() & 0xffff;
            Class clazz = messageClass[type];
            if (clazz == null) {
                throw new MalformedP2PMessageException("unknown message type " + type);
            }
            P2PMessage msg;
            try {
                msg = (P2PMessage) clazz.newInstance();
            } catch (InstantiationException e) {
                throw new RuntimeException(e);
            } catch (IllegalAccessException e) {
                throw new RuntimeException(e);
            }
            msg.deserialize(buffer);
            return msg;
        }

    }

}